/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * ZXSimulator: a prototype ZX80/81 BASIC Simulator
 */


#include stdio_h


/****************************** cmd_h *******************************/

#define ZXBREAK   ((keyrow(7)==2) && (keyrow(3)==8))       /* ctl-s */
#define getexpr    evalbool    /* top expression in precedence tree */

#define addr       long

#define EOL       '\n'
#define EOP       '\0'
#define LINELEN    128  /* was 256 */
#define CMDLEN     20
#define STRLEN     80
#define VARLEN     20
#define BLKLEN     64
#define COLLEN     32
#define NUMLEN     10
#define MAXVAR     7     /* needs to be prime (i.e. 257) */
#define MAXLNUM    9999

#define MAXNEST    10
#define NOLP       1
#define FORLP      2

#define NOERR      0
#define NOCMD     -1
#define NOVAL     -2
#define NOHEAP    -3
#define MAXLIM    -4
#define BADARG    -5
#define BADCHAR   -6
#define NOQUOTE   -7
#define UNKNOWN   -8

/* constants for our random number generator */
#define P1         8421
#define P2         1
#define N          65537

/* constants for ZX graphics */
#define TOPROW     1
#define MAXCOL     262
#define MAXROW     188
#define FIRSTCOL   2
#define LASTCOL    250
#define MIDCOL     130
#define FIRSTROW   2
#define LASTROW    170
#define CURSROW    180

/* global command line variables */
int  globerr = NOERR;
char pre_ws[5] = {' ','(',')','\n','\0'};
char pos_ws[6] = {'-',' ','(',')','\n','\0'};
char arg_dl[3] = {'\"','\'','\0'};
char cstr[LINELEN] = "";
char *pstr;
int  plen = 0;
char *str;
int  sp = 0;

/* data sturctures for variables */
addr var_nam[MAXVAR] = {0L,0L,0L,0L,0L,0L,0L};
addr var_str[MAXVAR] = {0L,0L,0L,0L,0L,0L,0L};
int  var_num[MAXVAR] = {0,0,0,0,0,0,0};

/* data structures for loops gosubs, and if statements */
int  lp_ptr[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
int  lp_end[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
int  lp_inc[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
addr lp_var[MAXNEST] = {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L};
int  lp_ind = -1;

int  rt_ptr[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
addr rt_str[MAXNEST] = {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L};
int  rt_ind = -1;

/* random seed */
long seed;

/* data structurs for ZX graphics */
int  col = FIRSTCOL;
int  row = FIRSTROW;
int  endrow = LASTROW;
int  dquote = 0;
int  inprow;



/****************************** cmd_c *******************************/

main ()
{
   char cmd[CMDLEN];
   int ln;

   initPARSE();

   while (1)
   {
      reset(); globerr = NOERR;

      ZXgets(cstr, 0);

      if (linenum(&ln) == NOERR)  /* add new line */
      {
         if (addline(ln) != NOERR)
            showerr("BAD LINE NUMBER");
         else
            { findline(ln); listpgm(1); }
      }
      else 
      {
         if (getcmd(cmd) != NOERR) goto no_cmd;

         if (cmd[0]=='\0')   /* list program on blank line */
            listpgm(0);
         else
         {
            switch (PARSEcmds(cmd)) /* otherwise start parsing */
            {
               case NOERR  :
               case MAXLIM : break;
               case NOCMD  : no_cmd: showerr("NO COMMAND"); break;
               case BADARG : showerr("BAD ARGUMENT"); break;
               default     : showerr("UNKNOWN");
            }
         }
      }
   }
}   /* end main */


/* initializes the parser */
initPARSE ()
{
   initQLscr();
   seed = date();

   str = cstr;
   plen = 0;
   setpstr(0,EOL); setpstr(1,EOP);
}   /* end initPARSE */


/* parser */
PARSEcmds(cmd)
   char cmd[];
{
   int  err, ln;

   while (1)
   {
      if (ZXBREAK) return NOERR;

      if ((err=BASICcmds(cmd)) != NOERR) return err;

      if (str[sp] == EOL)        /* go to next line */
         { sp++; linenum(&ln); }
      if (str[sp] == EOP) return NOERR;
      if (getcmd(cmd) != NOERR) return BADARG;
   }
}   /* end PARSEcmds */


/* BASIC commands */
BASICcmds (cmd)
   char cmd[];
{
   char *val_s, *vl2_s, *ostr, arg[LINELEN], var[VARLEN];
   int  val_i, vl2_i, vl3_i, err, op, i, doLF;

   if (streql("CLS",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      cls(stdout);
      row = FIRSTROW; col = FIRSTCOL;
   }
   else if (streql("EDIT",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (editline(val_i) != NOERR) return BADARG;
   }
   else if (streql("FOR",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (getsym("=") != NOERR) return BADARG;
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("TO") != NOERR) return BADARG;
      if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("STEP") != NOERR)
         vl3_i = 1;
      else if (getexpr(&vl3_i,&val_s) == NOERR)
         free(val_s);
      else
         return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (nestlp(var,val_i,vl2_i,vl3_i) != NOERR) return BADARG;
   }
   else if (streql("GOSUB",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if ((++rt_ind) >= MAXNEST) return BADARG;
      rt_ptr[rt_ind] = sp; rt_str[rt_ind] = str;
      findline(val_i);
   }
   else if (streql("GOTO",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      findline(val_i);
   }
   else if (streql("IF",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("THEN") != NOERR) return BADARG;
      if (noarg() == NOERR) return BADARG;
      if (!val_i) skipln();
   }
   else if (streql("INPUT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;

      ZXgets(arg,1); arg[strlen(arg)-1]='\0';
      storevar(var,atoi(arg),arg);
   }
   else if (streql("LET",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (getsym("=") != NOERR)
         { fetchvar(var,&val_i,&val_s);
           if (getrange(&vl2_i,&vl3_i,strlen(val_s)) != NOERR) { free(val_s); return BADARG; }
           if (getsym("=") != NOERR) { free(val_s); return BADARG; }
           if (getexpr(&val_i,&vl2_s) != NOERR) { free(val_s); return BADARG; }
           for (i=vl2_i-1; i<vl3_i; i++)
              if (vl2_s[i-vl2_i+1]=='\0') break; else val_s[i]=vl2_s[i-vl2_i+1];
           val_i=atoi(val_s); free(vl2_s); }
      else
         { if (getexpr(&val_i,&val_s) != NOERR) return BADARG; }

      if (noarg() != NOERR) { free(val_s); return BADARG; }
      storevar(var,val_i,val_s); free(val_s);
   }
   else if (streql("LIST",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           op = sp; ostr = str; findline(val_i); }
      else
         { op = sp; ostr = str; sp = 1; }

      if (listpgm(0) != NOERR) return MAXLIM;
      sp = op; str = ostr;
   }
   else if (streql("LOAD",cmd))
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if (noarg() != NOERR) { free(val_s); return BADARG; }
      if (loadfile(val_s) != NOERR) { free(val_s); return BADARG; }
      free(val_s);
   }
   else if (streql("LRUN",cmd))
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if (noarg() != NOERR)
         { if (getsym(",") != NOERR) { free(val_s); return BADARG; }
           if (getexpr(&vl2_i,&vl2_s) == NOERR) free(vl2_s); else { free(val_s); return BADARG; }
           if (noarg() != NOERR) { free(val_s); return BADARG; }
           if (loadfile(val_s) != NOERR) { free(val_s); return BADARG; }
           free(val_s); reset(); findline(val_i); }
      else
         { if (loadfile(val_s) != NOERR) {free(val_s); return BADARG; }
           free(val_s); reset(); str = pstr; }
   }
   else if (streql("NEW",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      reset(); setpstr(0,EOL); setpstr(1,EOP);
   }
   else if (streql("NEXT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (nextlp(var) != NOERR) return BADARG;
   }
   else if (streql("PRINT",cmd))
   {
      doLF = 1;
      if (getkey("AT") == NOERR)
          { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
            if (getsym(",") != NOERR) return BADARG;
            if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
            row = (FIRSTROW + val_i * 8); if (row < 0 || row > LASTROW) return BADARG;
            col = (FIRSTCOL + vl2_i * 8); if (col < 0 || col > LASTCOL) return BADARG;
            if (row == LASTROW) doLF = 0; }
      if ((err=putformat(doLF)) != NOERR) return err;
   }
   else if (streql("RAND",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           seed = val_i; }
      else
         { seed = date(); }
   }
   else if (streql("REM",cmd))
   {
      skipln();
   }
   else if (streql("RETURN",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      if (rt_ind < 0) return BADARG;
      sp = rt_ptr[rt_ind]; str = rt_str[rt_ind--];
   }
   else if (streql("RUN",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           reset(); findline(val_i); }
      else
         { reset(); str = pstr; }
   }
   else if (streql("SAVE",cmd))
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if (noarg() != NOERR) { free(val_s); return BADARG; }
      if (savefile(val_s) != NOERR) return BADARG;
      free(val_s);
   }
   else if (streql("STOP",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      reset();
   }
   else if (streql("QUIT",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      abort(0);
   }
   else if (cmd[0] != '\0')        /* if command is not empty */
      return NOCMD;

   return NOERR;
}   /* end BASICcmds */


/****************************** aux_c *******************************/

/* gets command from command line */
getcmd (cmd)
   char cmd[];
{
   char ch;
   int  i;

   while (str[sp] == ' ') sp++;
   for (i=0; i<CMDLEN; i++,sp++)
      if (((ch=str[sp]) != ' ') && (ch != '\n'))
         cmd[i] = toupper(ch);
      else
        { cmd[i] = '\0'; return NOERR; }
   return BADARG;
}   /* end getcmd */


/* gets symbol from command line */
getsym (sym)
   char sym[];
{
   int i, op;

   op = sp;
   while (str[sp] == ' ') sp++;
   for (i=0; sym[i]!='\0'; i++)
      if (toupper(str[sp]) == sym[i]) sp++; else { sp=op; return BADCHAR; }
   return NOERR;
}   /* getsym */


/* gets keyword from command line */
getkey (key)
   char key[];
{
   char ch;
   int i, tp;

   tp = sp;
   while (str[tp] == ' ') tp++;
   if (strchr(pre_ws,str[tp-1]) == NULL) return BADCHAR;
   for (i=0; key[i]!='\0'; i++)
      if (toupper(str[tp]) == key[i]) tp++; else return BADCHAR;
   if (strchr(pos_ws,str[tp]) == NULL) return BADCHAR;
   sp = tp;
   return NOERR;
}   /* getkey */


/* finds keyword from command line */
findkey (key, ech)
   char key[], ech;
{
   char ch;

   ch = key[0];

   while (str[sp] != ech)
   {
      if (toupper(str[sp]) == ch) if (getkey(key) == NOERR) return NOERR;
      sp++;
   }

   return BADARG;
}   /* findkey */


/* finds keyword from command line within range bp & ep */
instrkey (key, bp, ep)
   char key[];
   int  bp, ep;
{
   char ch;

   ch = key[0];

   for (sp=bp; sp<ep; sp++)
      if (toupper(str[sp]) == ch) if (getkey(key) == NOERR) return NOERR;

   return BADARG;
}   /* instrkey */


/* displays formatted argument for print statement */
putformat (lf)
   int lf;
{
   int i, val_i, err, doLF;
   char *val_s;

   err = NOERR; doLF = lf;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case ','  : { sp++; doLF=0; if (col<MIDCOL) col=MIDCOL; else err=downcursor(); break; }
         case ';'  : { sp++; doLF=0; break; }
         case '\n' : { if (doLF) return ZXprintf("\n"); return NOERR; }
         default   : {
                       if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
                       doLF = lf; err=ZXprintf(val_s); free(val_s);
                     }
      }
      if (err != NOERR) return err;
   }
}   /* end putformat */


/* gets next argument from command line */
nextarg (arg)
   char arg[];
{
   int i;
   char dl, ch;

   while (str[sp] == ' ') sp++;
   if (str[sp-1] != ' ') return BADARG;
   if (strchr(arg_dl,str[sp]) == NULL) dl = ' '; else dl = str[sp++];

   for (i=0; i<LINELEN; i++,sp++)
      if (((ch=str[sp])!=dl) && (ch!='\n') && (dl!=' '))
         arg[i] = ch;
      else
         { arg[i] = '\0';
           if (dl != ' ') if (dl == ch) sp++; else return NOQUOTE;
           if (i) return NOERR; else return NOVAL; }

   return BADARG;
}   /* end nextarg */


/* gets last argument from command line */
lastarg (arg)
   char arg[];
{
   int err;

   if ((err=nextarg(arg)) != NOERR) return err;
   else if ((err=noarg()) != NOERR) return err;
   else return NOERR;
}   /* end lastarg */


/* checks for no argument from command line */
noarg ()
{
   while (str[sp] == ' ') sp++;
   if (str[sp] != '\n')
      return BADCHAR;
   else
      return NOERR;
}   /* end noarg */


/* skips to end of line */
skipln ()
{
   while (str[sp] != '\n') sp++;
}   /* end skipln */


/* clears upto "st" looping variables */
clearlp (st)
   int st;
{
   int i;

   for (i=st; i<=lp_ind; i++)
      { lp_ptr[i]=0; lp_end[i]=0; lp_inc[i]=0; free(lp_var[i]); lp_var[i]=0L; }
   lp_ind = st-1;
}   /* end clearlp */


/* skips to next increment of "var" */
nextlp (var)
   char var[];
{
   addr val_s;
   int i, val_i, inc, end;

   if (var[strlen(var)-1] == '$') return BADARG;

   for (i=lp_ind; i>=0; i--)
      if (streql(lp_var[i],var))
      {
         fetchvar(var,&val_i,&val_s); free(val_s);
         inc = lp_inc[i];  end = lp_end[i];
         val_i += inc;
         if (((inc > 0) && (val_i > end)) || ((inc < 0) && (val_i < end)))
            { clearlp(i); }
         else
            { storevar(var,val_i,""); clearlp(i+1); sp=lp_ptr[i]; }
         return NOERR;
      }
   return BADARG;
}   /* end nextlp */


/* adds new loop nest for "var" */
nestlp (var,bgn,end,inc)
   char var[];
   int bgn, end, inc;
{
   addr mem;
   int  l;

   if ((lp_ind+1) == MAXNEST) return BADARG;
   if (var[strlen(var)-1] == '$') return BADARG;

   storevar(var,bgn,"");

   if (((inc > 0) && (bgn > end)) || ((inc < 0) && (bgn < end)))
      return BADARG;

   if (lp_var[++lp_ind] != 0L) free(lp_var[lp_ind]);
   if ((mem=malloc(strlen(var)+1)) == NULL)
      syserr(NOHEAP);
   else
   {
      lp_ptr[lp_ind] = sp;
      lp_end[lp_ind] = end;
      lp_inc[lp_ind] = inc;
      lp_var[lp_ind] = mem; strcpy(lp_var[lp_ind],var);
   }
      
   return NOERR;
}   /* end nestlp */


/* list program */
listpgm (ed)
   int ed;
{
   int i, j, k, cp;
   char lstr[LINELEN];

   dquote = 1; /* show double quote: "" */

   cp = sp;
   for (i=0; i<=(ed*10); i++)  /* if editing, new line to middle of screen */
      if (sp > 1) { sp--; while (pstr[sp-1]!=EOL) sp--; } else sp=1;

   j = 0;
   for (i=sp; pstr[i]!=EOP; i++)
   {
      if (j == 0)
      {
         if (pstr[i+1] == ' ')
            { lstr[j++] = ' '; goto in2; }  /* reduce code to indent 3 21*/
         else if (pstr[i+2] == ' ')
            { in2: lstr[j++] = ' '; goto in3; } /* reduce code to indent 2  */
         else if (pstr[i+3] == ' ')
            { in3: lstr[j++] = ' '; }
      }

      if (ed && i == cp)
         { lstr[j++] = '%'; lstr[j++] = '>'; }
      else
         lstr[j++] = pstr[i];
         
      if (pstr[i] == EOL)
      {
         lstr[j] = '\0';
         if ((ZXlen(lstr) % COLLEN) == 1) lstr[j-1] = '\0';
         if (ZXprintf(lstr) != NOERR) { dquote = 0; return MAXLIM; }
         j = 0;
      }
   }

   dquote = 0; /* show single quote: " */
   return NOERR;
}   /* end listpgm */


/* saves program to file */
savefile (fname)
   char fname[];
{
   int fp, i;

   for (i=0; fname[i]!='\0'; i++) fname[i] = tolower(fname[i]); /* lowcase name */

   if ((fp=fopen(fname,"w")) == NULL) return BADARG;
   for (i=1; pstr[i]!=EOP; i++)
   {
      fputc(pstr[i],fp);

      /* we change our intneral \"" to \" to be compatible */
      if (pstr[i-1] == '\\' && pstr[i] == '"' && pstr[i+1] == '"') i++;
   }
       
   fclose(fp);

   return NOERR;
}   /* end savefile */


/* loads file into string */
loadfile (fname)
   char fname[];
{
   int fp, l, ol, ln, ch, err;

   if ((fp=fopen(fname,"r")) == NULL) return BADARG;
   l = 0; ln = 0; setpstr(0,'\n');
   while ((err=checkline(fp,&ln,&l)) == NOERR)
      if ((err=readline(fp,&l)) != NOERR) break;

   if (err != BADCHAR)
      { if (pstr[l] != '\n') setpstr(++l,'\n'); setpstr(++l,'\0'); }
   else
      { ol=l; readline(fp,&l); pgmerr("BAD LINE",ol); setpstr(1,'\0'); }
   fclose(fp);
   return NOERR;
}   /* end loadfile */


/* reads character into program string, realloc's memory */
setpstr (l,ch)
   int l;
   char ch;
{
   char *ostr, *pmem;
   int  i;

   if (l >= plen)
   {
      if ((pmem=malloc(l+BLKLEN)) == NULL) { syserr(NOHEAP); return; }
      ostr = pstr; pstr = pmem;
      for (i=0; i<l; i++) pstr[i] = ostr[i];
      if (plen) free(ostr);
      plen = l + BLKLEN;
   }
   pstr[l] = ch;
}    /* end setpstr */


/* checks line number from file */
checkline (fp,ln,l)
   int fp, *ln, *l;
{
   char lnum[NUMLEN];
   int  i, num, ch, err;

   i = 0;
   while (((ch=fgetc(fp)) == ' ') || (ch == '\n'));
   if (ch == EOF) return EOF;
   while (isdigit(ch))
   {
      lnum[i++] = ch; setpstr(++(*l),ch);
      if ((ch=fgetc(fp)) == EOF) { (*l) -= i; return EOF; }
   }

   lnum[i] = '\0'; num = atoi(lnum); setpstr(++(*l),ch);
   if ((ch == ' ') && (num > *ln) && (num <= MAXLNUM))
      { *ln = num; return NOERR; }
   else
      return BADCHAR;
}   /* checkline */


/* reads line from file */
readline (fp,l)
   int fp, *l;
{
   int ch;

   if (pstr[*l] != '\n')
   {
      while ((ch=toupper(fgetc(fp))) != '\n')
      {
         if (ch == '\\')     /* we change \" to our intneral \"" to be compatible */
            { setpstr(++(*l),ch); if ((ch=toupper(fgetc(fp))) == '"') setpstr(++(*l),ch); }
         if (ch == EOF) return EOF; else setpstr(++(*l),ch);
      }
      setpstr(++(*l),'\n');
   }
   return NOERR;
}   /* readline */


/* finds line number or gives error */
findline (num)
   int num;
{
   int lnum;

   sp = 0; str = pstr;
   while (str[sp] != EOP)
      if (str[sp++] == EOL)
         if (linenum(&lnum) == NOERR)
         {
            if (num < lnum)
               return NOVAL;
            else if (num == lnum)
               return NOERR;
         }
   return NOVAL;
}   /* findline */


/* gets line from program */
getline (lp, bp, ep)
   int lp, *bp, *ep;
{
   char ch;

   *bp=lp; while (*bp) if (pstr[(*bp)-1] != EOL) (*bp)--; else break;
   *ep=lp; while ((ch=pstr[*ep]) != EOP) if (ch != EOL) (*ep)++; else break;
}   /* getline */


/* gets line number from command line */
linenum (ln)
   int * ln;
{
   char num[NUMLEN];
   int  i;

   i=0;
   while(str[sp] == ' ') sp++;
   while(isdigit(str[sp])) num[i++]=str[sp++];
   num[i] = '\0'; *ln = atoi(num);
   if (i) return NOERR; else return NOVAL;
}   /* end linenum */


/* adds new line to program */
addline (num)
   int num;
{
   int err, cl, bp, cp, ep, np;

   if ((num < 1) || (num > MAXLNUM)) return BADARG;
   cp = sp; while (cstr[cp] == ' ') cp++;

   /* first delete line if found */
   err = findline(num); getline(sp,&bp,&ep);
   if (err == NOERR)
      { np=bp; while (pstr[ep] != EOP) pstr[np++] = pstr[++ep]; }
   else
      ep = strlen(pstr);

   /* if new line is not empty, add it */
   if (cstr[cp] != EOL)
   {
      np=0; while ((cstr[np]==' ') || (cstr[np]=='0')) np++;
      cl=strlen(cstr)-np; while (ep >= bp) setpstr(ep+cl, pstr[ep--]);
      for (cp=np; cstr[cp]!='\0'; cp++) setpstr(bp++, cstr[cp]);
   }

   return NOERR;
}   /* end addline */


/* edits line of program */
editline (num)
   int num;
{
   int  i, ln, bp, ep;

   if (findline(num) == NOERR)
   {
      getline(sp,&bp,&ep);
      for (i=bp; i<ep; i++) cstr[i-bp] = pstr[i];
      cstr[ep-bp] = '\0';

      ZXgets(cstr, 0);
      sp = 0; str = cstr;
      if (linenum(&num) == NOERR)
      {
         if (addline(num) != NOERR)  return BADARG;
         listpgm(1);
      }
      reset(); return NOERR;
   }
   return NOVAL;
}   /* end editline */


/* points to error in program line */
pgmerr (msg,lp)
   char msg[];
   int  lp;
{
   char estr[LINELEN];
   int  i, bp, ep;

   getline(lp,&bp,&ep); for (i=bp; i<ep; i++) estr[i-bp] = pstr[i];
   estr[ep-bp] = '\n'; estr[ep-bp+1] = '\0';
   printerr(msg,estr,lp-bp);
}   /* end pgmerr */


/* points to error in command line */
showerr (msg)
   char msg[];
{
   if (str == cstr)
      printerr(msg,"",sp);
   else
      pgmerr(msg,sp);
   reset();
}   /* end showerr */


/* displays error message */
printerr (msg,estr,ep)
   char msg[], estr[];
   int  ep;
{
   int i, pt;

   if (estr[0] == '\0')
   {
      row = FIRSTROW; col = FIRSTCOL;
   }
   else
   {
      ZXprintf(estr);

      ep = ep % COLLEN;
      if (ep > (COLLEN/2)) pt=ep-strlen(msg)-3; else pt=ep;
      for(i=0; i<pt; i++)
         ZXprintf(" ");
      
      ZXprintf("<__");
   }
   ZXprintf(msg);
}   /* end printerr */


/* dynamic memory allocation of arg */
memarg (arg,len)
   addr *arg; int len;
{
   addr mem;

   if (globerr == NOERR)
      if (len > 0)
      {
         if ((mem=malloc(len)) == NULL)
            syserr(NOHEAP);
         else
            *arg = mem;
      }
}   /* end memarg */


/* sets char ptr to ch */
setchp (chp,ind,ch)
   char chp[]; int ind; char ch;
{
   chp[ind] = ch;
}   /* end setchp */


/* gets ch from char ptr */
getchp (chp,ind)
   char chp[]; int ind;
{
   return chp[ind];
}   /* end getchp */


/* gives system error */
syserr (err)
   int err;
{
   int ew;

   if (globerr == NOERR)
   {
      switch(globerr=err)
      {
         case NOHEAP : showerr("NOT ENOUGH HEAP SPACE"); break;
         case MAXLIM : showerr("MAXIMUM LIMIT REACHED"); break;
         default     : showerr("UNKNOWN SYSTEM ERROR");
      }
      reset();
   }
}   /* end syserr */


/* reset input strings */
reset ()
{
   cstr[0] = '\0'; str = cstr; sp = 0;
   rt_ind = -1; clearlp(0);
}   /* end reset */


/* compares 2 strings */
streql (s,t)
   char *s, *t;
{
   while (*s == *t) { if (*s == '\0') return 1; s++; t++; }
   return 0;
}   /* end streql */


/* hash function */
hash (var)
    char var[];
{
   int i, h;

   h=0;
   for (i=0; var[i]!='\0'; i++) h+=var[i];

   return (h % MAXVAR);
}   /* end hash */


/*
 * QL specific commands begin here
 */


/* initializes the display */
initQLscr ()
{
   window(stdout,MAXCOL+4,MAXROW+4,(MAXCOL+4)/2,TOPROW);
   paper(stdout,7);
   ink(stdout,0);
   border(stdout,1,0);
   mode(4);
}   /* end initQLscr */


/* gets input ZX81 style */
ZXgets (lstr, io)
   char lstr[]; int io;
{
   int i, j, k, r, c, in;
   char ch, curs, cl[LINELEN];

   dquote = 1;        /* show double quote: "" */
   endrow = CURSROW;  /* set end row to edit mode */
   inprow = CURSROW;

   k=0; in=0; 
   if (io)
   {
      r = row; c = col; curs = 'L';
      cl[k++] = '%'; cl[k++] = curs;
   }
   else
   {
      curs = 'K';
      if (lstr[0] == '\0')
      {
         cl[k++] = '%'; cl[k++] = curs;
      }
      else
      {
         j=0;
         while (lstr[j]!='\0')
         {
            if (!in && lstr[j] == ' ')
               { in = k; cl[k++] = '%'; cl[k++] = curs; }
            
            cl[k++] = lstr[j++];
         }
      }
   }
   cl[k++] = ' ';  /* trailing space to remove prev char w/deleting */
   cl[k++] = '\0';
         
   for(;;)
   {
      ZXclear(cl);
      ZXprintf(cl);

      ch = toupper(getchar());

      if (ch == 200)
      {
         if (in < (k-6) && cl[in+2] == '\\')
            { cl[in] = '\\'; cl[in+1] = cl[in+3]; cl[in+2] = cl[in+4];
              cl[in+3] = '%'; cl[in+4] = curs; in = in + 3; }
         else if (in < (k-5) && cl[in+2] == '%')
            { cl[in+1] = cl[in+3]; cl[in+3] = curs; in = in + 2; }
         else if (in < (k-4))
            { cl[in] = cl[in+2]; cl[in+1] = '%'; cl[in+2] = curs; in = in + 1; }
      }
      else if (ch == 192)
      {
         if (in > 2 && cl[in-3] == '\\')
            { cl[in+1] = cl[in-1]; cl[in] = cl[in-2]; cl[in-1] = '\\';
              cl[in-3] = '%'; cl[in-2] = curs; in = in - 3; }
         else if (in > 1 && cl[in-2] == '%')
            { cl[in+1] = cl[in-1]; cl[in-1] = curs; in = in - 2; }
         else if (in > 0)
            { cl[in+1] = cl[in-1]; cl[in-1] = '%'; cl[in] = curs; in = in - 1; }
      }
      else if (ch == 194)
      {
         if (in > 2 && cl[in-3] == '\\')
            { for (i=in-3; i<(k-2); i++) cl[i] = cl[i+3]; in = in - 3; k = k - 3; }
         else if (in > 1 && cl[in-2] == '%')
            { for (i=in-2; i<(k-1); i++) cl[i] = cl[i+2]; in = in - 2; k = k - 2;}
         else if (in > 0)
            { for (i=in-1; i<k; i++) cl[i] = cl[i+1]; in = in - 1; k = k - 1; }
      }
      else if (ch == 10)
      {
         i=0; j=0;
         while (j < k)
         {
            if (j == in)
               j = j + 2;
            lstr[i++] = cl[j++];
         }
         lstr[i-2] = '\n'; /* replace trailing space with new-line */
                           /* lstr[i-1] = '\0'; don't think I need this */

         ZXclear(cl);
         break;
      }
      else 
      {
         if (ch != '\\' && ch != '%')
            { for (i=k; i>=(in+1); i--) cl[i] = cl[i-1]; cl[in++] = ch; k = k + 1; }
         else
         {
            cl[in+1] = 'G';
            ZXclear(cl);
            ZXprintf(cl);
           
            if (ch == '\\')
               { for (i=k+2; i>=(in+3); i--) cl[i] = cl[i-3]; cl[in++] = '\\';
                 cl[in++] = toupper(getchar()); cl[in++] = toupper(getchar());
                 k = k + 3; }
            else if (ch == '%')
               { for (i=k+1; i>=(in+2); i--) cl[i] = cl[i-2]; cl[in++] = '%'; 
                 cl[in++] = toupper(getchar()); k = k + 2; }

            cl[in+1] = curs;
         }
      }
   }

   dquote = 0;        /* back to single quote: "      */
   endrow = LASTROW;  /* reset end row to screen mode */
   if (!io)
   {
      row = FIRSTROW; col = FIRSTCOL;
      cls(stdout);
   }
   else
   {
      row = r; col = c;
   }
}


/* gets number of ZX characters */
ZXlen(lstr)
   char lstr[];
{
   int i, len;

   len = 0;
   for (i=0; lstr[i]!='\0'; i++)
      { switch(lstr[i]) { case '\\': i++; case '%': i++; default: len++; } }

   return len;
}


/* clears edit window */
ZXclear (lstr)
   char lstr[];
{
   int i, j, r;


   r = CURSROW - ((ZXlen(lstr) / COLLEN) * 8);
   if (r < inprow)
   {
      inprow = r;
      j = (1 + ((CURSROW - r) / 8)) * COLLEN;

      row = r; col = FIRSTCOL - 8;
      for (i=0; i<j; i++)
      {
         adjustcursor();
         ZXputc(' ', 0, 7);
      }
   }
   else
      r = inprow;

   row = r; col = FIRSTCOL;
}


/* prints in ZX81 font */
ZXprintf (lstr)
   char lstr[];
{
   int i;

   for (i=0; lstr[i]!='\0'; i++)
   {
      switch (lstr[i])
      {
         case '\n':
            if (downcursor() != NOERR) return MAXLIM;
            continue;
         case '\\':
            ZXgraphics(lstr[++i], lstr[++i]);
            break;
         case '%':
            ZXputc(lstr[++i], 7, 0);
            break;
         default:
            ZXputc(lstr[i], 0, 7);
      }
      if (adjustcursor() != NOERR) return MAXLIM;
   }
   return NOERR;
}


/* prints single ZX81 font character */
ZXputc (ch, ink, pap)
   char ch; int ink, pap;
{
   block(stdout, 8, 8, col, row, pap);

   switch (ch)
   {
     /* LETTERS */

      case 'A':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 5, col+1, row+2, ink);
         block(stdout, 1, 5, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         return;
      case 'B':
         block(stdout, 5, 6, col+1, row+1, ink);
         block(stdout, 4, 1, col+2, row+2, pap);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 4, 2, col+2, row+4, pap);
         block(stdout, 1, 2, col+6, row+4, ink);
         return;
/*->*/case 'G':
         block(stdout, 3, 1, col+4, row+4, ink);
      case 'C':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+6, row+5, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case 'D':
         block(stdout, 4, 1, col+1, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 2, col+6, row+3, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 4, 1, col+1, row+6, ink);
         return;
      case 'E':
         block(stdout, 5, 1, col+2, row+6, ink);
      case 'F':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 5, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         return;
/*->  case 'G': <-- moved before letter C */
      case 'H':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         block(stdout, 1, 6, col+6, row+1, ink);
         return;
      case 'I':
         block(stdout, 5, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+4, row+2, ink);
         block(stdout, 5, 1, col+2, row+6, ink);
         return;
/*->*/case 'U':
         block(stdout, 1, 3, col+1, row+1, ink);
      case 'J':  /* adds extra block call 3 vs 4 */
         block(stdout, 1, 2, col+1, row+4, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         block(stdout, 1, 5, col+6, row+1, ink);
         return;
      case 'K':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 1, 1, col+5, row+1, ink);
         block(stdout, 1, 1, col+4, row+2, ink);
         block(stdout, 2, 1, col+2, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 1, col+6, row+6, ink);
         return;
      case 'L':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 5, 1, col+2, row+6, ink);
         return;
      case 'M':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 1, 6, col+6, row+1, ink);
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 2, 1, col+3, row+3, ink);
         return;
      case 'N':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+3, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 6, col+6, row+1, ink);
         return;
/*->*/case '0': /* zero */
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
      case 'O':
         letter_O:
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 1, 4, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
/*->*/case 'R':
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 1, col+6, row+6, ink);
      case 'P':
         block(stdout, 5, 1, col+1, row+1, ink);
         block(stdout, 1, 5, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         block(stdout, 1, 2, col+6, row+2, ink);
         return;
      case 'Q':
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+4, row+5, ink);
         goto letter_O;
/*->  case 'R': <-- moved before letter P */
      case 'S':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         block(stdout, 1, 2, col+6, row+4, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case 'T':
         block(stdout, 7, 1, col, row+1, ink);
         block(stdout, 1, 5, col+3, row+2, ink);
         return;
/*->  case 'U': <-- moved before letter J */
      case 'V':
         block(stdout, 1, 4, col+1, row+1, ink);
         block(stdout, 1, 4, col+6, row+1, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 2, 1, col+3, row+6, ink);
         return;
      case 'W':
         block(stdout, 1, 5, col+1, row+1, ink);
         block(stdout, 1, 5, col+6, row+1, ink);
         block(stdout, 2, 1, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case 'X':
         block(stdout, 1, 1, col+1, row+1, ink);
         block(stdout, 1, 1, col+6, row+1, ink);
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 2, 2, col+3, row+3, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 1, col+1, row+6, ink);
         block(stdout, 1, 1, col+6, row+6, ink);
         return;
      case 'Y':
         block(stdout, 1, 1, col, row+1, ink);
         block(stdout, 1, 1, col+6, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+2, row+3, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 3, col+3, row+4, ink);
         return;
      case 'Z':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
         block(stdout, 6, 1, col+1, row+6, ink);
         return;

      /* NUMBERS */

/*->  case '0': <-- moved before letter O */
      case '1':
         block(stdout, 2, 1, col+3, row+1, ink);
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 4, col+4, row+2, ink);
         block(stdout, 5, 1, col+2, row+6, ink);
         return;
      case '2':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 2, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         block(stdout, 6, 1, col+1, row+6, ink);
         return;
      case '3':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 2, 1, col+4, row+3, ink);
         block(stdout, 1, 2, col+6, row+4, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case '4':
         block(stdout, 1, 6, col+4, row+1, ink);
         block(stdout, 1, 1, col+3, row+2, ink);
         block(stdout, 1, 1, col+2, row+3, ink);
         block(stdout, 1, 1, col+1, row+4, ink);
         block(stdout, 6, 1, col+1, row+5, ink);
         return;
      case '5':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 2, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         block(stdout, 1, 2, col+6, row+4, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case '6':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         block(stdout, 1, 2, col+6, row+4, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case '7':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+5, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 2, col+3, row+5, ink);
         return ;
      case '8':
         block(stdout, 4, 6, col+2, row+1, ink);
         block(stdout, 6, 4, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+2, pap);
         block(stdout, 4, 2, col+2, row+4, pap);
         block(stdout, 1, 1, col+1, row+3, pap);
         block(stdout, 1, 1, col+6, row+3, pap);
         return ;
      case '9':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 2, col+1, row+2, ink);
         block(stdout, 1, 4, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;

      /* SYMBOLS */

      case '"':
         block(stdout, 1, 2, col+3, row+1, ink);
         block(stdout, 1, 2, col+6, row+1, ink);
         return;
      case '#':
         block(stdout, 3, 1, col+3, row+1, ink);
         block(stdout, 1, 4, col+2, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 4, 1, col+1, row+3, ink);
         block(stdout, 6, 1, col+1, row+6, ink);
         return;
      case '$':
         block(stdout, 5, 5, col+2, row+2, ink);
         block(stdout, 4, 1, col+3, row+3, pap);
         block(stdout, 4, 1, col+2, row+5, pap);
         block(stdout, 1, 7, col+4, row+1, ink);
         return;
      case ':':
         block(stdout, 1, 1, col+3, row+3, ink);
         block(stdout, 1, 1, col+3, row+6, ink);
          return;
      case '(':
         block(stdout, 1, 1, col+5, row+1, ink);
         block(stdout, 1, 4, col+4, row+2, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case ')':
         block(stdout, 1, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+3, row+2, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         return;
      case '>':
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+3, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         return;
      case '<':
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+4, row+5, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case '=':
         block(stdout, 5, 1, col+2, row+3, ink);
         block(stdout, 5, 1, col+2, row+5, ink);
         return;
      case '+':
         block(stdout, 1, 5, col+4, row+2, ink);
      case '-':
         block(stdout, 5, 1, col+2, row+4, ink);
         return;
      case '_':
         block(stdout, 5, 1, col+2, row+6, ink);
         return;
      case '*':
         block(stdout, 1, 3, col+4, row+3, ink);
         block(stdout, 5, 1, col+2, row+4, ink);
         block(stdout, 1, 1, col+3, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+3, row+6, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case '/':
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+5, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         return;
      case ';':
         block(stdout, 1, 1, col+3, row+2, ink);
         block(stdout, 1, 2, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+7, ink);
         return;
      case ',':
         block(stdout, 1, 2, col+4, row+5, ink);
         block(stdout, 1, 1, col+3, row+7, ink);
         return;
      case '.':
         block(stdout, 2, 2, col+3, row+5, ink);
         return;
      case ' ':
         return;

      /* UNKOWN is ? */

      default:
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+5, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+4, row+6, ink);
         return;
   }
}


/* prints ZX81 graphics character */
ZXgraphics (ch1, ch2)
   char ch1, ch2;
{
   switch (ch1)
   {
      case '"':
         if (ch2 == '"')
            { ZXputc ('"', 0, 7);
              if (dquote)
                 { if (adjustcursor() == NOERR)
                    ZXputc ('"', 0, 7); }
              return; }
         break;
      case '#':
         if (ch2 == '#')
            { ZXgrey(1, 1, 0, 7); return; }
         break;
      case '~':
         if (ch2 == '~')
            { ZXgrey(1, 0, 0, 7); return; }
         break;
      case ',':
         if (ch2 == ',')
            { ZXgrey(0, 1, 0, 7); return; }
         break;
      case '@':
         if (ch2 == '@')
            { ZXgrey(1, 1, 7, 0); return; }
         break;
      case '!':
         if (ch2 == '!')
            { ZXgrey(1, 0, 7, 0); return; }
         break;
      case ';':
         if (ch2 == ';')
            { ZXgrey(0, 1, 7, 0); return; }
         break;
      case ' ':
         if (ch2 == ':')
            { ZXblock(0, 1, 0, 1); return; }
         if (ch2 == '\'')
            { ZXblock(0, 1, 0, 0); return; }
         if (ch2 == '.')
            { ZXblock(0, 0, 0, 1); return; }
         if (ch2 == ' ')
            { ZXblock(0, 0, 0, 0); return; }
               return;
         break;
      case ':':
         if (ch2 == ':')
            { ZXblock(1, 1, 1, 1); return; }
         if (ch2 == '\'')
            { ZXblock(1, 1, 1, 0); return; }
         if (ch2 == '.')
            { ZXblock(1, 0, 1, 1); return; }
         if (ch2 == ' ')
            { ZXblock(1, 0, 1, 0); return; }
         break;
      case '\'':
         if (ch2 == ':')
            { ZXblock(1, 1, 0, 1); return; }
         if (ch2 == '\'')
            { ZXblock(1, 1, 0, 0); return; }
         if (ch2 == '.')
            { ZXblock(1, 0, 0, 1); return; }
         if (ch2 == ' ')
            { ZXblock(1, 0, 0, 0); return; }
         break;
      case '.':
         if (ch2 == ':')
            { ZXblock(0, 1, 1, 1); return; }
         if (ch2 == '\'')
            { ZXblock(0, 1, 1, 0); return; }
         if (ch2 == '.')
            { ZXblock(0, 0, 1, 1); return; }
         if (ch2 == ' ')
            { ZXblock(0, 0, 1, 0); return; }
         break;
   }

   /* Question mark for unknown characters */
   ZXputc('?', 0, 7);
}


/* create specfic ZX81 grey character */
ZXgrey (up, lo, ink, pap)
   int up, lo, ink, pap;
{
   int pat;

   if (pap == 0)
      pat = 246;  /* even stipple pattern */
   else
      pat = 248;  /* odd stipple pattern */

   block(stdout, 8, 8, col, row, pap);

   if (up)
   {
      block(stdout, 8, 4, col, row, pat);
   }
   if (lo)
   {
      block(stdout, 8, 4, col, row+4, pat);
   }
}


/* create ZX81 block characters */
ZXblock (ul, ur, ll, lr)
   int ul, ur, ll, lr;
{
   block(stdout, 8, 8, col, row, 7);

   if (ul)
      block(stdout, 4, 4, col, row, 0);
   if (ur)
      block(stdout, 4, 4, col+4, row, 0);
   if (ll)
      block(stdout, 4, 4, col, row+4, 0);
   if (lr)
      block(stdout, 4, 4, col+4, row+4, 0);
}


/* move cursor down */
downcursor ()
{
   col = FIRSTCOL;
   row = row + 8;
   if (row > LASTROW)
   {
      row = FIRSTROW;
      return MAXLIM;
   }
   return NOERR;
}


/* adjust cursor for end of line/screen */
adjustcursor ()
{
   col = col + 8;
   if (col > LASTCOL)
   {
      col = FIRSTCOL;
      row = row + 8;
      if (row > endrow)
      {
         row = FIRSTROW;
         return MAXLIM;
      }
   }
   return NOERR;
}


/****************************** expr_c ******************************/


/* gets a variable from command line */
getvar (var)
   char var[];
{
   int  i;
   char ch;

   while (str[sp] == ' ') sp++;
   if (!isalpha(str[sp])) return NOVAL;
   for (i=0; i<VARLEN; i++,sp++)
      if (isalpha(ch=str[sp]) || isdigit(ch) || (ch == '_'))
         var[i] = toupper(ch);
      else
         { if (ch == '$')
              { var[i++] = ch; sp++; if (i == VARLEN) break; }
           var[i] = '\0'; return NOERR; }
   return BADARG;
}   /* end getvar */


/* finds var in hashtable */
findvar(var,vp)
   char var[]; int *vp;
{
   int i, j;

   i = hash(var);
   for (j=i; j!=(i-1); j=(j+1)%MAXVAR)
      if (var_nam[j] == 0L)
         { *vp = j; return NOVAL; }
      else if (streql(var_nam[j],var))
         { *vp = j; return NOERR; }

   return MAXLIM;
}   /* end findvar */


/* sets a variable */
storevar (var,val_i,val_s)
   char var[]; int val_i; char val_s[];
{
   addr mem;
   int  i, len, err;

   if ((err=findvar(var,&i)) == MAXLIM)
      syserr(err);
   else
   {
      if (err == NOVAL)
         { memarg(&mem,strlen(var)+1);
           var_nam[i]=mem; strcpy(mem,var); }

      if (var[strlen(var)-1] != '$')
         var_num[i] = val_i;
      else
         { memarg(&mem,strlen(val_s)+1);
           if (err == NOERR) free(var_str[i]);
           var_str[i] = mem; strcpy(mem,val_s); }
   }
}   /* end storevar */


/* retrievs value of a variable */
fetchvar (var,val_i,val_s)
   char var[]; int *val_i; addr *val_s;
{
   int i, err;

   if ((err=findvar(var,&i)) == MAXLIM)
      syserr(err);
   else if (err == NOVAL)
      { memarg(val_s,1);
        *val_i = 0; strcpy(*val_s,""); }
   else if (var[strlen(var)-1] != '$')
      { memarg(val_s,NUMLEN);
        *val_i = var_num[i]; itoa(*val_i,*val_s); }
   else
      { memarg(val_s,strlen(var_str[i])+1);
        strcpy(*val_s,var_str[i]); *val_i = atoi(*val_s); }
}   /* end fetchvar */


/* gets number from command line */
getnum (val_i,val_s)
   int *val_i; addr *val_s;
{
   int i;

   while (str[sp] == ' ') sp++;
/*
   if (!isdigit(str[sp]) && ((str[sp]!='-') || !isdigit(str[sp+1]))) return NOVAL;
*/
   if (!isdigit(str[sp])) return NOVAL;
   i=0; memarg(val_s,NUMLEN);
   if (str[sp] == '-') setchp(*val_s,i++,str[sp++]);
   while (isdigit(str[sp]))
      setchp(*val_s,i++,str[sp++]);
   setchp(*val_s,i,'\0');
   *val_i = atoi(*val_s);

   return NOERR;
}   /* end getnum */


/* gets quoted string from command line */
getstr (val_i,val_s)
   int *val_i; addr *val_s;
{
   int i, tp, len;

   while (str[sp] == ' ') sp++;
   if (str[sp] == '"') tp = ++sp; else return NOVAL;
   while(str[tp]!='"') { if (str[tp]=='\\') tp=tp+3; else if (str[tp++]=='\n') return BADCHAR; }
   len = tp-sp; memarg(val_s,len+1);
   for (i=0; i<len; i++) setchp(*val_s,i,str[sp++]);
   setchp(*val_s,i,'\0'); sp++;
   *val_i = atoi(*val_s);

   return NOERR;   
}   /* end getstr */


/* parses substring expression */
getsubstr (tmp_i,tmp_s,val_i,val_s)
   int *tmp_i; addr *tmp_s;
   int *val_i; addr *val_s;
{
   char ch;
   int err, from, to, i, j, cp, in, idx, len;

   if ((err=getrange(&from,&to,(len=strlen(*tmp_s)))) == NOVAL)
      { *val_i=*tmp_i; *val_s=*tmp_s; return NOERR; }
   else if (err != NOERR)
      { free(*tmp_s); return BADARG; }
   
   for (cp=0; cp<2; cp++)  /* first pass see how much space, no copy */
   {
      if (cp) memarg(val_s,j);

      i=0; j=0; in=0; idx=0;
      while (i<len)
      {
         switch (ch=getchp(*tmp_s,i))
         {
            case '%':
               idx++;
               if (in || from == idx) 
               {
                  if (cp) 
                     { setchp(*val_s,j++, ch); i++;
                       setchp(*val_s,j++,getchp(*tmp_s,i)); i++; }
                  else
                     { i=i+2; j=j+2; }
                  if (idx == to) goto end_idx;
                  in = 1; 
               }
               else
                  i=i+2;
               break;
            case '\\':
               idx++;
               if (in || from == idx)
               {
                  if (cp) 
                     { setchp(*val_s,j++, ch); i++;
                       setchp(*val_s,j++,getchp(*tmp_s,i)); i++;
                       setchp(*val_s,j++,getchp(*tmp_s,i)); i++; }
                  else
                     { i=i+3; j=j+3; }
                  if (idx == to) goto end_idx;
                  in = 1; 
               }
               else
                  i=i+3;
               break;
            default:
               idx++;
               if (in || from == idx)
               {
                  if (cp) setchp(*val_s,j,ch);
                  j++;
                  if (idx == to) goto end_idx;
                  in = 1; 
               }
               i++;
         }
      }
      end_idx:
   }
   setchp(*val_s,j,'\0');  *val_i = atoi(*val_s);

   free(*tmp_s);
   return NOERR;
}   /* end getsubstr */


/* gets range of string */
getrange (from,to,len)
   int *from, *to, len;
{
   char *tmp;
   int  noTO;

   if (getsym("(") != NOERR) return NOVAL;

   if (noTO=(getkey("TO") != NOERR))
      { if (getexpr(from,&tmp) ==NOERR) free(tmp); else return BADARG; }
   else
      *from = 1;

   if (!noTO || (getkey("TO") == NOERR))
      { if (getexpr(to,&tmp) == NOERR) free(tmp); else *to = len; }
   else
      *to = *from;

   if (getsym(")") != NOERR) return BADARG;

   if (*from < 1) *from = 1;
   if (*to > len) *to = len;
   if (*to < *from) *to = *from-1;
   return NOERR;
}   /* end getrange */


/* gets value from command line */
getval (val_i,val_s)
   int *val_i; addr *val_s;
{
   addr tmp_s;
   int  tmp_i, err;
   char var[VARLEN];

   if (getsym("(") == NOERR)  /* eval subexpression */
      { if (getexpr(val_i,val_s) != NOERR) return BADARG;
        return getsym(")"); }
   else if ((err=evalfun(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getnum(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getstr(&tmp_i,&tmp_s)) != NOVAL)
      { if (err != NOERR) return err;
        return getsubstr(&tmp_i,&tmp_s,val_i,val_s); }
   else if (getvar(var) == NOERR)
      { fetchvar(var,&tmp_i,&tmp_s);
        return getsubstr(&tmp_i,&tmp_s,val_i,val_s); }

   return BADARG;
}   /* end getval */


/* gets {chr$(),len()} functions and unary {-,not} expressions */
evalfun (val_i,val_s)
   int *val_i; addr *val_s;
{
   while (str[sp] == ' ') sp++;
   if (getsym("-") == NOERR) {
      if (getval(val_i,val_s) != NOERR) return BADARG;
      *val_i = -*val_i;
   }
   else if (getsym("CHR$") == NOERR) {
      if (getexpr(val_i,val_s) != NOERR) return BADARG;
      *val_i %= 256; setchp(*val_s,0,*val_i); if (*val_i) setchp(*val_s,1,'\0');
      return NOERR;
   }
   else if (getsym("CODE") == NOERR) {
      if (getexpr(val_i,val_s) != NOERR) return BADARG;
      *val_i = getchp(*val_s,0);
   }
   else if (getsym("LEN") == NOERR) {
      if (getexpr(val_i,val_s) != NOERR) return BADARG;
      *val_i = strlen(*val_s);
   }
   else if (getsym("RND") == NOERR) {
      seed = ((seed * P1) + P2) % N; *val_i = seed;
   }
   else if (getkey("NOT") == NOERR) {
      if (evaleql(val_i,val_s) != NOERR) return BADARG;
      *val_i = !*val_i;
   }
   else return NOVAL;

   free(*val_s); memarg(val_s,NUMLEN); itoa(*val_i,*val_s);
   return NOERR;
}   /* end evalfun */


/* parses ** expressions */
evalpow (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i, pow, i;
   addr right_s;

   if (getval(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            if (str[sp+1] != '*') return NOERR;
            sp = sp + 2;
            if (getval(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            pow = 1;
            for (i=0; i<right_i; i++)
               pow *= *left_i;

            *left_i = pow;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free (*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evalpow */


/* parses {*,/} expressions */
evalmul (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evalpow(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i * right_i;
            break;
         }
         case '/' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i / right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evalmul */


/* parses {+,-} expressions */
evaladd (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evalmul(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '+' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i + right_i;
            break;
         }
         case '-' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i - right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evaladd */


/* parses {=,>,<,<>,<=,
} expressions */
evaleql (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evaladd(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getsym("=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i == right_i;
         else
            *left_i = (strcmp(*left_s,right_s) == 0);
      }
      else if (getsym("<=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i <= right_i;
         else
            *left_i = (strcmp(*left_s,right_s) <= 0);
      }
      else if (getsym("<>") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i != right_i;
         else
            *left_i = (strcmp(*left_s,right_s) != 0);
      }
      else if (getsym("<") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i < right_i;
         else
            *left_i = (strcmp(*left_s,right_s) < 0);
      }
      else if (getsym(">=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i >= right_i;
         else
            *left_i = (strcmp(*left_s,right_s) >= 0);
      }
      else if (getsym(">") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i > right_i;
         else
            *left_i = (strcmp(*left_s,right_s) > 0);
      }
      else {
         return NOERR;
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evaleql */


/* parses {and,or} expressions */
evalbool (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evaleql(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getkey("AND") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         *left_i = *left_i && right_i;
      }
      else if (getkey("OR") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         *left_i = *left_i || right_i;
      }
      else {
         return NOERR;
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evalbool */

